@@ -6,13 +6,17 @@
 -- > import qualified Data.OpenPGP as OpenPGP
 module Data.OpenPGP (
 	Packet(
+		AsymmetricSessionKeyPacket,
 		OnePassSignaturePacket,
+		SymmetricSessionKeyPacket,
 		PublicKeyPacket,
 		SecretKeyPacket,
 		CompressedDataPacket,
 		MarkerPacket,
 		LiteralDataPacket,
+		TrustPacket,
 		UserIDPacket,
+		EncryptedDataPacket,
 		ModificationDetectionCodePacket,
 		UnsupportedPacket,
 		compression_algorithm,
@@ -30,15 +34,11 @@ module Data.OpenPGP (
 		key_id,
 		message,
 		nested,
-		private_hash,
-		s2k_count,
-		s2k_hash_algorithm,
-		s2k_salt,
-		s2k_type,
 		s2k_useage,
+		s2k,
 		signature,
 		signature_type,
-		symmetric_type,
+		symmetric_algorithm,
 		timestamp,
 		trailer,
 		unhashed_subpackets,
@@ -48,6 +48,8 @@ module Data.OpenPGP (
 	signaturePacket,
 	Message(..),
 	SignatureSubpacket(..),
+	S2K(..),
+	string2key,
 	HashAlgorithm(..),
 	KeyAlgorithm(..),
 	SymmetricAlgorithm(..),
@@ -56,28 +58,33 @@ module Data.OpenPGP (
 	MPI(..),
 	find_key,
 	fingerprint_material,
-	signatures_and_data,
-	signature_issuer
+	SignatureOver(..),
+	signatures,
+	signature_issuer,
+	public_key_fields,
+	secret_key_fields
 ) where
 
 import Numeric
 import Control.Monad
-import Control.Exception (assert)
+import Control.Arrow
+import Control.Applicative
+import Data.Monoid
 import Data.Bits
 import Data.Word
 import Data.Char
-import Data.Maybe
 import Data.List
 import Data.OpenPGP.Internal
+import qualified Data.ByteString as BS
 import qualified Data.ByteString.Lazy as LZ
 
 #ifdef CEREAL
-import Data.Serialize
+import Data.Serialize hiding (decode)
 import qualified Data.ByteString as B
 import qualified Data.ByteString.UTF8 as B (toString, fromString)
 #define BINARY_CLASS Serialize
 #else
-import Data.Binary
+import Data.Binary hiding (decode)
 import Data.Binary.Get
 import Data.Binary.Put
 import qualified Data.ByteString.Lazy as B
@@ -99,8 +106,10 @@ getSomeByteString = getByteString . fromIntegral
 putSomeByteString :: B.ByteString -> Put
 putSomeByteString = putByteString
 
-unsafeRunGet :: Get a -> B.ByteString -> a
-unsafeRunGet g bs = let Right v = runGet g bs in v
+localGet :: Get a -> B.ByteString -> Get a
+localGet g bs = case runGet g bs of
+	Left s -> fail s
+	Right v -> return v
 
 compress :: CompressionAlgorithm -> B.ByteString -> B.ByteString
 compress algo = toStrictBS . lazyCompress algo . toLazyBS
@@ -113,6 +122,9 @@ toStrictBS = B.concat . LZ.toChunks
 
 toLazyBS :: B.ByteString -> LZ.ByteString
 toLazyBS = LZ.fromChunks . (:[])
+
+lazyEncode :: (Serialize a) => a -> LZ.ByteString
+lazyEncode = toLazyBS . encode
 #else
 getRemainingByteString :: Get B.ByteString
 getRemainingByteString = getRemainingLazyByteString
@@ -123,14 +135,21 @@ getSomeByteString = getLazyByteString . fromIntegral
 putSomeByteString :: B.ByteString -> Put
 putSomeByteString = putLazyByteString
 
-unsafeRunGet :: Get a -> B.ByteString -> a
-unsafeRunGet = runGet
+localGet :: Get a -> B.ByteString -> Get a
+localGet g bs = case runGetOrFail g bs of
+	Left (_,_,s) -> fail s
+	Right (leftover,_,v)
+		| B.null leftover -> return v
+		| otherwise -> fail $ "Leftover in localGet: " ++ show leftover
 
 compress :: CompressionAlgorithm -> B.ByteString -> B.ByteString
 compress = lazyCompress
 
 decompress :: CompressionAlgorithm -> B.ByteString -> B.ByteString
 decompress = lazyDecompress
+
+lazyEncode :: (Binary a) => a -> LZ.ByteString
+lazyEncode = encode
 #endif
 
 lazyCompress :: CompressionAlgorithm -> LZ.ByteString -> LZ.ByteString
@@ -147,10 +166,29 @@ lazyDecompress ZLIB         = Zlib.decompress
 lazyDecompress BZip2        = BZip2.decompress
 lazyDecompress x            = error ("No implementation for " ++ show x)
 
-assertProp :: (a -> Bool) -> a -> a
-assertProp f x = assert (f x) x
+assertProp :: (Monad m, Show a) => (a -> Bool) -> a -> m a
+assertProp f x
+	| f x = return $! x
+	| otherwise = fail $ "Assertion failed for: " ++ show x
+
+pad :: Int -> String -> String
+pad l s = replicate (l - length s) '0' ++ s
+
+padBS :: Int -> B.ByteString -> B.ByteString
+padBS l s = B.replicate (fromIntegral l - B.length s) 0 `B.append` s
+
+checksum :: B.ByteString -> Word16
+checksum = fromIntegral .
+	B.foldl (\c i -> (c + fromIntegral i) `mod` 65536) (0::Integer)
 
 data Packet =
+	AsymmetricSessionKeyPacket {
+		version::Word8,
+		key_id::String,
+		key_algorithm::KeyAlgorithm,
+		encrypted_data::B.ByteString
+	} |
+	-- ^ <http://tools.ietf.org/html/rfc4880#section-5.1>
 	SignaturePacket {
 		version::Word8,
 		signature_type::Word8,
@@ -162,6 +200,14 @@ data Packet =
 		signature::[MPI],
 		trailer::B.ByteString
 	} |
+	-- ^ <http://tools.ietf.org/html/rfc4880#section-5.2>
+	SymmetricSessionKeyPacket {
+		version::Word8,
+		symmetric_algorithm::SymmetricAlgorithm,
+		s2k::S2K,
+		encrypted_data::B.ByteString
+	} |
+	-- ^ <http://tools.ietf.org/html/rfc4880#section-5.3>
 	OnePassSignaturePacket {
 		version::Word8,
 		signature_type::Word8,
@@ -170,6 +216,7 @@ data Packet =
 		key_id::String,
 		nested::Word8
 	} |
+	-- ^ <http://tools.ietf.org/html/rfc4880#section-5.4>
 	PublicKeyPacket {
 		version::Word8,
 		timestamp::Word32,
@@ -178,34 +225,41 @@ data Packet =
 		is_subkey::Bool,
 		v3_days_of_validity::Maybe Word16
 	} |
+	-- ^ <http://tools.ietf.org/html/rfc4880#section-5.5.1.1> (also subkey)
 	SecretKeyPacket {
 		version::Word8,
 		timestamp::Word32,
 		key_algorithm::KeyAlgorithm,
 		key::[(Char,MPI)],
-		s2k_useage::Word8, -- determines if the Maybes are Just or Nothing
-		symmetric_type::Maybe Word8,
-		s2k_type::Maybe Word8,
-		s2k_hash_algorithm::Maybe HashAlgorithm,
-		s2k_salt::Maybe Word64,
-		s2k_count::Maybe Word32,
+		s2k_useage::Word8,
+		s2k::S2K, -- ^ This is meaningless if symmetric_algorithm == Unencrypted
+		symmetric_algorithm::SymmetricAlgorithm,
 		encrypted_data::B.ByteString,
-		private_hash::Maybe B.ByteString, -- the hash may be in the encrypted data
 		is_subkey::Bool
 	} |
+	-- ^ <http://tools.ietf.org/html/rfc4880#section-5.5.1.3> (also subkey)
 	CompressedDataPacket {
 		compression_algorithm::CompressionAlgorithm,
 		message::Message
 	} |
-	MarkerPacket |
+	-- ^ <http://tools.ietf.org/html/rfc4880#section-5.6>
+	MarkerPacket | -- ^ <http://tools.ietf.org/html/rfc4880#section-5.8>
 	LiteralDataPacket {
 		format::Char,
 		filename::String,
 		timestamp::Word32,
 		content::B.ByteString
 	} |
-	UserIDPacket String |
-	ModificationDetectionCodePacket B.ByteString |
+	-- ^ <http://tools.ietf.org/html/rfc4880#section-5.9>
+	TrustPacket B.ByteString | -- ^ <http://tools.ietf.org/html/rfc4880#section-5.10>
+	UserIDPacket String | -- ^ <http://tools.ietf.org/html/rfc4880#section-5.11>
+	EncryptedDataPacket {
+		version::Word8,
+		encrypted_data::B.ByteString
+	} |
+	-- ^ <http://tools.ietf.org/html/rfc4880#section-5.13>
+	-- or <http://tools.ietf.org/html/rfc4880#section-5.7> when version is 0
+	ModificationDetectionCodePacket B.ByteString | -- ^ <http://tools.ietf.org/html/rfc4880#section-5.14>
 	UnsupportedPacket Word8 B.ByteString
 	deriving (Show, Read, Eq)
 
@@ -214,56 +268,63 @@ instance BINARY_CLASS Packet where
 		-- First two bits are 1 for new packet format
 		put ((tag .|. 0xC0) :: Word8)
 		case tag of
-			19 -> put (assertProp (<192) blen :: Word8)
+			19 -> put =<< assertProp (<192) (blen :: Word8)
 			_  -> do
 				-- Use 5-octet lengths
 				put (255 :: Word8)
 				put (blen :: Word32)
-				putSomeByteString body
+		putSomeByteString body
 		where
 		blen :: (Num a) => a
 		blen = fromIntegral $ B.length body
 		(body, tag) = put_packet p
 	get = do
-		tag <- get :: Get Word8
+		tag <- get
 		let (t, l) =
 			if (tag .&. 64) /= 0 then
 				(tag .&. 63, parse_new_length)
 			else
-				((tag `shiftR` 2) .&. 15, parse_old_length tag)
-		len <- l
-		-- This forces the whole packet to be consumed
-		packet <- getSomeByteString (fromIntegral len)
-		return $ unsafeRunGet (parse_packet t) packet
+				((tag `shiftR` 2) .&. 15, (,) <$> parse_old_length tag <*> pure False)
+		packet <- uncurry get_packet_bytes =<< l
+		localGet (parse_packet t) (B.concat packet)
+
+get_packet_bytes :: Maybe Word32 -> Bool -> Get [B.ByteString]
+get_packet_bytes len partial = do
+	-- This forces the whole packet to be consumed
+	packet <- maybe getRemainingByteString (getSomeByteString . fromIntegral) len
+	if not partial then return [packet] else
+		(packet:) <$> (uncurry get_packet_bytes =<< parse_new_length)
 
 -- http://tools.ietf.org/html/rfc4880#section-4.2.2
-parse_new_length :: Get Word32
-parse_new_length = do
+parse_new_length :: Get (Maybe Word32, Bool)
+parse_new_length = fmap (first Just) $ do
 	len <- fmap fromIntegral (get :: Get Word8)
 	case len of
 		-- One octet length
-		_ | len < 192 -> return len
+		_ | len < 192 -> return (len, False)
 		-- Two octet length
 		_ | len > 191 && len < 224 -> do
 			second <- fmap fromIntegral (get :: Get Word8)
-			return $ ((len - 192) `shiftL` 8) + second + 192
+			return (((len - 192) `shiftL` 8) + second + 192, False)
 		-- Five octet length
-		255 -> get :: Get Word32
-		-- TODO: Partial body lengths. 1 << (len & 0x1F)
+		255 -> (,) <$> (get :: Get Word32) <*> pure False
+		-- Partial length (streaming)
+		_ | len >= 224 && len < 255 ->
+			return (1 `shiftL` (fromIntegral len .&. 0x1F), True)
 		_ -> fail "Unsupported new packet length."
 
 -- http://tools.ietf.org/html/rfc4880#section-4.2.1
-parse_old_length :: Word8 -> Get Word32
+parse_old_length :: Word8 -> Get (Maybe Word32)
 parse_old_length tag =
 	case tag .&. 3 of
 		-- One octet length
-		0 -> fmap fromIntegral (get :: Get Word8)
+		0 -> fmap (Just . fromIntegral) (get :: Get Word8)
 		-- Two octet length
-		1 -> fmap fromIntegral (get :: Get Word16)
+		1 -> fmap (Just . fromIntegral) (get :: Get Word16)
 		-- Four octet length
-		2 -> get
+		2 -> fmap Just get
 		-- Indeterminate length
-		3 -> fmap fromIntegral remaining
+		3 -> return Nothing
 		-- Error
 		_ -> fail "Unsupported old packet length."
 
@@ -286,7 +347,7 @@ secret_key_fields DSA     = ['x']
 secret_key_fields _       = undefined -- Nothing in the spec. Maybe empty
 
 (!) :: (Eq k) => [(k,v)] -> k -> v
-(!) xs = fromJust . (`lookup` xs)
+(!) xs k = let Just x = lookup k xs in x
 
 -- Need this seperate for trailer calculation
 signature_packet_start :: Packet -> B.ByteString
@@ -334,7 +395,14 @@ calculate_signature_trailer p@(SignaturePacket {version = 4}) =
 calculate_signature_trailer x =
 	error ("Trying to calculate signature trailer for: " ++ show x)
 
-put_packet :: (Num a) => Packet -> (B.ByteString, a)
+put_packet :: Packet -> (B.ByteString, Word8)
+put_packet (AsymmetricSessionKeyPacket version key_id key_algorithm dta) =
+	(B.concat [
+		encode version,
+		encode (fst $ head $ readHex $ takeFromEnd 16 key_id :: Word64),
+		encode key_algorithm,
+		dta
+	], 1)
 put_packet (SignaturePacket { version = v,
                               unhashed_subpackets = unhashed_subpackets,
                               key_algorithm = key_algorithm,
@@ -353,10 +421,12 @@ put_packet (SignaturePacket { version = v,
 		encode hash_head
 	] ++ map encode signature, 2)
 	where
-	keyid = fst $ head $ readHex keyidS :: Word64
+	keyid = fst $ head $ readHex $ takeFromEnd 16 keyidS :: Word64
 	Just (IssuerPacket keyidS) = find isIssuer unhashed_subpackets
 	isIssuer (IssuerPacket {}) = True
 	isIssuer _ = False
+put_packet (SymmetricSessionKeyPacket version salgo s2k encd) =
+	(B.concat [encode version, encode salgo, encode s2k, encd], 3)
 put_packet (SignaturePacket { version = 4,
                               unhashed_subpackets = unhashed_subpackets,
                               hash_head = hash_head,
@@ -376,44 +446,34 @@ put_packet (OnePassSignaturePacket { version = version,
                                      key_algorithm = key_algorithm,
                                      key_id = key_id,
                                      nested = nested }) =
-	(B.concat [ encode version, encode signature_type,
-	             encode hash_algorithm, encode key_algorithm,
-	             encode (fst $ head $ readHex key_id :: Word64),
-	             encode nested ], 4)
+	(B.concat [
+		encode version, encode signature_type,
+		encode hash_algorithm, encode key_algorithm,
+		encode (fst $ head $ readHex $ takeFromEnd 16 key_id :: Word64),
+		encode nested
+	], 4)
 put_packet (SecretKeyPacket { version = version, timestamp = timestamp,
                               key_algorithm = algorithm, key = key,
-                              s2k_useage = s2k_useage,
-                              symmetric_type = symmetric_type,
-                              s2k_type = s2k_type,
-                              s2k_hash_algorithm = s2k_hash_algo,
-                              s2k_salt = s2k_salt,
-                              s2k_count = s2k_count,
+                              s2k_useage = s2k_useage, s2k = s2k,
+                              symmetric_algorithm = symmetric_algorithm,
                               encrypted_data = encrypted_data,
                               is_subkey = is_subkey }) =
-	(B.concat $ [p, encode s2k_useage] ++
-	(if s2k_useage `elem` [255, 254] then
-		[encode $ fromJust symmetric_type, encode s2k_t,
-		 encode $ fromJust s2k_hash_algo] ++
-		(if s2k_t `elem` [1,3] then [encode $ fromJust s2k_salt] else []) ++
-		if s2k_t == 3 then
-			[encode $ encode_s2k_count $ fromJust s2k_count] else []
-	else []) ++
-	(if s2k_useage > 0 then
+	(B.concat $ p :
+	(if s2k_useage `elem` [254,255] then
+		[encode s2k_useage, encode symmetric_algorithm, encode s2k]
+	else
+		[encode symmetric_algorithm]
+	) ++
+	(if symmetric_algorithm /= Unencrypted then
+		-- For V3 keys, the "encrypted data" has an unencrypted checksum
+		-- of the unencrypted MPIs on the end
 		[encrypted_data]
 	else s ++
-		-- XXX: Checksum is part of encrypted_data for V4 ONLY
-		if s2k_useage == 254 then
-			[B.replicate 20 0] -- TODO SHA1 Checksum
-		else
-			[encode (fromIntegral $
-				B.foldl (\c i -> (c + fromIntegral i) `mod` 65536)
-				(0::Integer) (B.concat s) :: Word16)]),
+		[encode $ checksum $ B.concat s]),
 	if is_subkey then 7 else 5)
 	where
-	(Just s2k_t) = s2k_type
 	p = fst (put_packet $
-		PublicKeyPacket version timestamp algorithm key False Nothing
-		:: (B.ByteString, Integer)) -- Supress warning
+		PublicKeyPacket version timestamp algorithm key False Nothing)
 	s = map (encode . (key !)) (secret_key_fields algorithm)
 put_packet p@(PublicKeyPacket { version = v, timestamp = timestamp,
                               key_algorithm = algorithm, key = key,
@@ -421,7 +481,7 @@ put_packet p@(PublicKeyPacket { version = v, timestamp = timestamp,
 	| v == 3 =
 		final (B.concat $ [
 			B.singleton 3, encode timestamp,
-			encode (fromJust $ v3_days_of_validity p),
+			encode v3_days,
 			encode algorithm
 		] ++ material)
 	| v == 4 =
@@ -429,6 +489,7 @@ put_packet p@(PublicKeyPacket { version = v, timestamp = timestamp,
 			B.singleton 4, encode timestamp, encode algorithm
 		] ++ material)
 	where
+	Just v3_days = v3_days_of_validity p
 	final x = (x, if is_subkey then 14 else 6)
 	material = map (encode . (key !)) (public_key_fields algorithm)
 put_packet (CompressedDataPacket { compression_algorithm = algorithm,
@@ -438,23 +499,35 @@ put_packet MarkerPacket = (B.fromString "PGP", 10)
 put_packet (LiteralDataPacket { format = format, filename = filename,
                                 timestamp = timestamp, content = content
                               }) =
-	(B.concat [encode format, encode filename_l, lz_filename,
-	            encode timestamp, content], 11)
+	(B.concat [
+		encode format, encode filename_l, lz_filename,
+		encode timestamp, content
+	], 11)
 	where
 	filename_l  = (fromIntegral $ B.length lz_filename) :: Word8
 	lz_filename = B.fromString filename
+put_packet (TrustPacket bytes) = (bytes, 12)
 put_packet (UserIDPacket txt) = (B.fromString txt, 13)
+put_packet (EncryptedDataPacket 0 encrypted_data) = (encrypted_data, 9)
+put_packet (EncryptedDataPacket version encrypted_data) =
+	(B.concat [encode version, encrypted_data], 18)
 put_packet (ModificationDetectionCodePacket bstr) = (bstr, 19)
 put_packet (UnsupportedPacket tag bytes) = (bytes, fromIntegral tag)
 put_packet x = error ("Unsupported Packet version or type in put_packet: " ++ show x)
 
 parse_packet :: Word8 -> Get Packet
+-- AsymmetricSessionKeyPacket, http://tools.ietf.org/html/rfc4880#section-5.1
+parse_packet  1 = AsymmetricSessionKeyPacket
+	<$> (assertProp (==3) =<< get)
+	<*> fmap (pad 16 . map toUpper . flip showHex "") (get :: Get Word64)
+	<*> get
+	<*> getRemainingByteString
 -- SignaturePacket, http://tools.ietf.org/html/rfc4880#section-5.2
 parse_packet  2 = do
 	version <- get
 	case version of
 		_ | version `elem` [2,3] -> do
-			_ <- fmap (assertProp (==5)) (get :: Get Word8)
+			_ <- assertProp (==5) =<< (get :: Get Word8)
 			signature_type <- get
 			creation_time <- get :: Get Word32
 			keyid <- get :: Get Word64
@@ -470,7 +543,7 @@ parse_packet  2 = do
 				hashed_subpackets = [],
 				unhashed_subpackets = [
 					SignatureCreationTimePacket creation_time,
-					IssuerPacket $ pad $ map toUpper $ showHex keyid ""
+					IssuerPacket $ pad 16 $ map toUpper $ showHex keyid ""
 				],
 				hash_head = hash_head,
 				signature = signature,
@@ -482,10 +555,10 @@ parse_packet  2 = do
 			hash_algorithm <- get
 			hashed_size <- fmap fromIntegral (get :: Get Word16)
 			hashed_data <- getSomeByteString hashed_size
-			let hashed = unsafeRunGet listUntilEnd hashed_data
+			hashed <- localGet listUntilEnd hashed_data
 			unhashed_size <- fmap fromIntegral (get :: Get Word16)
 			unhashed_data <- getSomeByteString unhashed_size
-			let unhashed = unsafeRunGet listUntilEnd unhashed_data
+			unhashed <- localGet listUntilEnd unhashed_data
 			hash_head <- get
 			signature <- listUntilEnd
 			return SignaturePacket {
@@ -500,8 +573,12 @@ parse_packet  2 = do
 				trailer = B.concat [encode version, encode signature_type, encode key_algorithm, encode hash_algorithm, encode (fromIntegral hashed_size :: Word16), hashed_data, B.pack [4, 0xff], encode ((6 + fromIntegral hashed_size) :: Word32)]
 			}
 		x -> fail $ "Unknown SignaturePacket version " ++ show x ++ "."
-	where
-	pad s = replicate (16 - length s) '0' ++ s
+-- SymmetricSessionKeyPacket, http://tools.ietf.org/html/rfc4880#section-5.3
+parse_packet  3 = SymmetricSessionKeyPacket
+	<$> (assertProp (==4) =<< get)
+	<*> get
+	<*> get
+	<*> getRemainingByteString
 -- OnePassSignaturePacket, http://tools.ietf.org/html/rfc4880#section-5.4
 parse_packet  4 = do
 	version <- get
@@ -515,11 +592,9 @@ parse_packet  4 = do
 		signature_type = signature_type,
 		hash_algorithm = hash_algo,
 		key_algorithm = key_algo,
-		key_id = pad $ map toUpper $ showHex key_id "",
+		key_id = pad 16 $ map toUpper $ showHex key_id "",
 		nested = nested
 	}
-	where
-	pad s = replicate (16 - length s) '0' ++ s
 -- SecretKeyPacket, http://tools.ietf.org/html/rfc4880#section-5.5.3
 parse_packet  5 = do
 	-- Parse PublicKey part
@@ -531,31 +606,24 @@ parse_packet  5 = do
 	}) <- parse_packet 6
 	s2k_useage <- get :: Get Word8
 	let k = SecretKeyPacket version timestamp algorithm key s2k_useage
-	k' <- case s2k_useage of
-		_ | s2k_useage `elem` [255, 254] -> do
-			symmetric_type <- get
-			s2k_type <- get
-			s2k_hash_algorithm <- get
-			s2k_salt <- if s2k_type `elem` [1, 3] then get
-				else return undefined
-			s2k_count <- if s2k_type == 3 then fmap decode_s2k_count get else
-				return undefined
-			return (k (Just symmetric_type) (Just s2k_type)
-				(Just s2k_hash_algorithm) (Just s2k_salt) (Just s2k_count))
+	(symmetric_algorithm, s2k) <- case () of
+		_ | s2k_useage `elem` [255, 254] -> (,) <$> get <*> get
 		_ | s2k_useage > 0 ->
 			-- s2k_useage is symmetric_type in this case
-			return (k (Just s2k_useage) Nothing Nothing Nothing Nothing)
+			(,) <$> localGet get (encode s2k_useage) <*> pure (SimpleS2K MD5)
 		_ ->
-			return (k Nothing Nothing Nothing Nothing Nothing)
-	if s2k_useage > 0 then do {
+			return (Unencrypted, S2K 100 B.empty)
+	if symmetric_algorithm /= Unencrypted then do {
 		encrypted <- getRemainingByteString;
-		return (k' encrypted Nothing False)
+		return (k s2k symmetric_algorithm encrypted False)
 	} else do
-		key <- foldM (\m f -> do
+		skey <- foldM (\m f -> do
 			mpi <- get :: Get MPI
-			return $ (f,mpi):m) key (secret_key_fields algorithm)
-		private_hash <- getRemainingByteString
-		return ((k' B.empty (Just private_hash) False) {key = key})
+			return $ (f,mpi):m) [] (secret_key_fields algorithm)
+		chk <- get
+		when (checksum (B.concat $ map (encode . snd) skey) /= chk) $
+			fail "Checksum verification failed for unencrypted secret key"
+		return ((k s2k symmetric_algorithm B.empty False) {key = key ++ skey})
 -- PublicKeyPacket, http://tools.ietf.org/html/rfc4880#section-5.5.2
 parse_packet  6 = do
 	version <- get :: Get Word8
@@ -593,11 +661,13 @@ parse_packet  7 = do
 -- CompressedDataPacket, http://tools.ietf.org/html/rfc4880#section-5.6
 parse_packet  8 = do
 	algorithm <- get
-	message <- getRemainingByteString
+	message <- localGet get =<< (decompress algorithm <$> getRemainingByteString)
 	return CompressedDataPacket {
 		compression_algorithm = algorithm,
-		message = unsafeRunGet get (decompress algorithm message)
+		message = message
 	}
+-- EncryptedDataPacket, http://tools.ietf.org/html/rfc4880#section-5.7
+parse_packet  9 = EncryptedDataPacket 0 <$> getRemainingByteString
 -- MarkerPacket, http://tools.ietf.org/html/rfc4880#section-5.8
 parse_packet 10 = return MarkerPacket
 -- LiteralDataPacket, http://tools.ietf.org/html/rfc4880#section-5.9
@@ -613,6 +683,8 @@ parse_packet 11 = do
 		timestamp = timestamp,
 		content = content
 	}
+-- TrustPacket, http://tools.ietf.org/html/rfc4880#section-5.10
+parse_packet 12 = fmap TrustPacket getRemainingByteString
 -- UserIDPacket, http://tools.ietf.org/html/rfc4880#section-5.11
 parse_packet 13 =
 	fmap (UserIDPacket . B.toString) getRemainingByteString
@@ -620,6 +692,8 @@ parse_packet 13 =
 parse_packet 14 = do
 	p <- parse_packet 6
 	return p {is_subkey = True}
+-- EncryptedDataPacket, http://tools.ietf.org/html/rfc4880#section-5.13
+parse_packet 18 = EncryptedDataPacket <$> get <*> getRemainingByteString
 -- ModificationDetectionCodePacket, http://tools.ietf.org/html/rfc4880#section-5.14
 parse_packet 19 =
 	fmap ModificationDetectionCodePacket getRemainingByteString
@@ -651,6 +725,48 @@ enum_to_word8 = fromIntegral . fromEnum
 enum_from_word8 :: (Enum a) => Word8 -> a
 enum_from_word8 = toEnum . fromIntegral
 
+data S2K =
+	SimpleS2K HashAlgorithm |
+	SaltedS2K HashAlgorithm Word64 |
+	IteratedSaltedS2K HashAlgorithm Word64 Word32 |
+	S2K Word8 B.ByteString
+	deriving (Show, Read, Eq)
+
+instance BINARY_CLASS S2K where
+	put (SimpleS2K halgo) = put (0::Word8) >> put halgo
+	put (SaltedS2K halgo salt) = put (1::Word8) >> put halgo >> put salt
+	put (IteratedSaltedS2K halgo salt count) = put (3::Word8) >> put halgo
+		>> put salt >> put (encode_s2k_count count)
+	put (S2K t body) = put t >> putSomeByteString body
+
+	get = do
+		t <- get :: Get Word8
+		case t of
+			0 -> SimpleS2K <$> get
+			1 -> SaltedS2K <$> get <*> get
+			3 -> IteratedSaltedS2K <$> get <*> get <*> (decode_s2k_count <$> get)
+			_ -> S2K t <$> getRemainingByteString
+
+-- | Take a hash function and an 'S2K' value and generate the bytes
+--   needed for creating a symmetric key.
+--
+-- Return value is always infinite length.
+-- Take the first n bytes you need for your keysize.
+string2key :: (HashAlgorithm -> LZ.ByteString -> BS.ByteString) -> S2K -> LZ.ByteString -> LZ.ByteString
+string2key hsh (SimpleS2K halgo) s = infiniHashes (hsh halgo) s
+string2key hsh (SaltedS2K halgo salt) s =
+	infiniHashes (hsh halgo) (lazyEncode salt `LZ.append` s)
+string2key hsh (IteratedSaltedS2K halgo salt count) s =
+	infiniHashes (hsh halgo) $
+	LZ.take (max (fromIntegral count) (LZ.length s))
+	(LZ.cycle $ lazyEncode salt `LZ.append` s)
+string2key _ s2k _ = error $ "Unsupported S2K specifier: " ++ show s2k
+
+infiniHashes :: (LZ.ByteString -> BS.ByteString) -> LZ.ByteString -> LZ.ByteString
+infiniHashes hsh s = LZ.fromChunks (hs 0)
+	where
+	hs c = hsh (LZ.replicate c 0 `LZ.append` s) : hs (c+1)
+
 data HashAlgorithm = MD5 | SHA1 | RIPEMD160 | SHA256 | SHA384 | SHA512 | SHA224 | HashAlgorithm Word8
 	deriving (Show, Read, Eq)
 
@@ -771,38 +887,99 @@ instance BINARY_CLASS RevocationCode where
 	put = put . enum_to_word8
 	get = fmap enum_from_word8 get
 
--- A message is encoded as a list that takes the entire file
+-- | A message is encoded as a list that takes the entire file
 newtype Message = Message [Packet] deriving (Show, Read, Eq)
 instance BINARY_CLASS Message where
 	put (Message xs) = mapM_ put xs
 	get = fmap Message listUntilEnd
 
--- | Extract all signature and data packets from a 'Message'
-signatures_and_data :: Message -> ([Packet], [Packet])
-signatures_and_data (Message ((CompressedDataPacket {message = m}):_)) =
-	signatures_and_data m
-signatures_and_data (Message lst) =
-	(filter isSignaturePacket lst, filter isDta lst)
+instance Monoid Message where
+	mempty = Message []
+	mappend (Message a) (Message b) = Message (a ++ b)
+
+-- | Data needed to verify a signature
+data SignatureOver =
+	DataSignature {literal::Packet, signatures_over::[Packet]} |
+	KeySignature {topkey::Packet, signatures_over::[Packet]} |
+	SubkeySignature {topkey::Packet, subkey::Packet, signatures_over::[Packet]} |
+	CertificationSignature {topkey::Packet, user_id::Packet, signatures_over::[Packet]}
+	deriving (Show, Read, Eq)
+
+-- To get the signed-over bytes
+instance BINARY_CLASS SignatureOver where
+	put (DataSignature (LiteralDataPacket {content = c}) _) =
+		putSomeByteString c
+	put (KeySignature k _) = mapM_ putSomeByteString (fingerprint_material k)
+	put (SubkeySignature k s _) = mapM_ (mapM_ putSomeByteString)
+		[fingerprint_material k, fingerprint_material s]
+	put (CertificationSignature k (UserIDPacket s) _) =
+		mapM_ (mapM_ putSomeByteString) [fingerprint_material k, [
+			B.singleton 0xB4,
+			encode ((fromIntegral $ B.length bs) :: Word32),
+			bs
+		]]
+		where
+		bs = B.fromString s
+	put x = fail $ "Malformed signature: " ++ show x
+	get = fail "Cannot meaningfully parse bytes to be signed over."
+
+-- | Extract signed objects from a well-formatted message
+--
+-- Recurses into CompressedDataPacket
+--
+-- <http://tools.ietf.org/html/rfc4880#section-11>
+signatures :: Message -> [SignatureOver]
+signatures (Message [CompressedDataPacket _ m]) = signatures m
+signatures (Message ps) =
+	maybe (paired_sigs Nothing ps) (\p -> [DataSignature p sigs]) (find isDta ps)
 	where
+	sigs = filter isSignaturePacket ps
 	isDta (LiteralDataPacket {}) = True
 	isDta _ = False
 
+-- TODO: UserAttribute
+paired_sigs :: Maybe Packet -> [Packet] -> [SignatureOver]
+paired_sigs _ [] = []
+paired_sigs _ (p@(PublicKeyPacket {is_subkey = False}):ps) =
+	KeySignature p (takeWhile isSignaturePacket ps) :
+	paired_sigs (Just p) (dropWhile isSignaturePacket ps)
+paired_sigs _ (p@(SecretKeyPacket {is_subkey = False}):ps) =
+	KeySignature p (takeWhile isSignaturePacket ps) :
+	paired_sigs (Just p) (dropWhile isSignaturePacket ps)
+paired_sigs (Just k) (p@(PublicKeyPacket {is_subkey = True}):ps) =
+	SubkeySignature k p (takeWhile isSignaturePacket ps) :
+	paired_sigs (Just p) (dropWhile isSignaturePacket ps)
+paired_sigs (Just k) (p@(SecretKeyPacket {is_subkey = True}):ps) =
+	SubkeySignature k p (takeWhile isSignaturePacket ps) :
+	paired_sigs (Just p) (dropWhile isSignaturePacket ps)
+paired_sigs (Just k) (p@(UserIDPacket {}):ps) =
+	CertificationSignature k p (takeWhile isSignaturePacket ps) :
+	paired_sigs (Just k) (dropWhile isSignaturePacket ps)
+paired_sigs k (_:ps) = paired_sigs k ps
+
+-- | <http://tools.ietf.org/html/rfc4880#section-3.2>
 newtype MPI = MPI Integer deriving (Show, Read, Eq, Ord)
 instance BINARY_CLASS MPI where
-	put (MPI i) = do
-		put (((fromIntegral . B.length $ bytes) - 1) * 8
-			+ floor (logBase (2::Double) $ fromIntegral (bytes `B.index` 0))
-			+ 1 :: Word16)
-		putSomeByteString bytes
+	put (MPI i)
+		| i >= 0 = do
+			put (bitl :: Word16)
+			putSomeByteString bytes
+		| otherwise = fail $ "MPI is less than 0: " ++ show i
 		where
-		bytes = if B.null bytes' then B.singleton 0 else bytes'
+		(bytes, bitl)
+			| B.null bytes' = (B.singleton 0, 1)
+			| otherwise     =
+				(bytes', (fromIntegral (B.length bytes') - 1) * 8 + sigBit)
+
+		sigBit = fst $ until ((==0) . snd)
+			(first (+1) . second (`shiftR` 1)) (0,B.index bytes 0)
 		bytes' = B.reverse $ B.unfoldr (\x ->
 				if x == 0 then Nothing else
 					Just (fromIntegral x, x `shiftR` 8)
-			) (assertProp (>=0) i)
+			) i
 	get = do
 		length <- fmap fromIntegral (get :: Get Word16)
-		bytes <- getSomeByteString ((length + 7) `div` 8)
+		bytes <- getSomeByteString =<< assertProp (>0) ((length + 7) `div` 8)
 		return (MPI (B.foldl (\a b ->
 			a `shiftL` 8 .|. fromIntegral b) 0 bytes))
 
@@ -814,15 +991,15 @@ listUntilEnd = do
 		rest <- listUntilEnd
 		return (next:rest)
 
--- http://tools.ietf.org/html/rfc4880#section-5.2.3.1
+-- | <http://tools.ietf.org/html/rfc4880#section-5.2.3.1>
 data SignatureSubpacket =
 	SignatureCreationTimePacket Word32 |
-	SignatureExpirationTimePacket Word32 | -- seconds after CreationTime
+	SignatureExpirationTimePacket Word32 | -- ^ seconds after CreationTime
 	ExportableCertificationPacket Bool |
 	TrustSignaturePacket {depth::Word8, trust::Word8} |
 	RegularExpressionPacket String |
 	RevocablePacket Bool |
-	KeyExpirationTimePacket Word32 | -- seconds after key CreationTime
+	KeyExpirationTimePacket Word32 | -- ^ seconds after key CreationTime
 	PreferredSymmetricAlgorithmsPacket [SymmetricAlgorithm] |
 	RevocationKeyPacket {
 		sensitive::Bool,
@@ -874,9 +1051,9 @@ instance BINARY_CLASS SignatureSubpacket where
 	get = do
 		len <- fmap fromIntegral (get :: Get Word8)
 		len <- case len of
-			_ | len > 190 && len < 255 -> do -- Two octet length
+			_ | len >= 192 && len < 255 -> do -- Two octet length
 				second <- fmap fromIntegral (get :: Get Word8)
-				return $ ((len - 192) `shiftR` 8) + second + 192
+				return $ ((len - 192) `shiftL` 8) + second + 192
 			255 -> -- Five octet length
 				fmap fromIntegral (get :: Get Word32)
 			_ -> -- One octet length, no furthur processing
@@ -884,7 +1061,7 @@ instance BINARY_CLASS SignatureSubpacket where
 		tag <- fmap stripCrit get :: Get Word8
 		-- This forces the whole packet to be consumed
 		packet <- getSomeByteString (len-1)
-		return $ unsafeRunGet (parse_signature_subpacket tag) packet
+		localGet (parse_signature_subpacket tag) packet
 		where
 		-- TODO: Decide how to actually encode the "is critical" data
 		-- instead of just ignoring it
@@ -911,10 +1088,10 @@ put_signature_subpacket (RevocationKeyPacket sensitive kalgo fpr) =
 	(B.concat [encode bitfield, encode kalgo, fprb], 12)
 	where
 	bitfield = 0x80 .|. (if sensitive then 0x40 else 0x0) :: Word8
-	fprb = B.drop 2 $ encode (MPI fpri)
+	fprb = padBS 20 $ B.drop 2 $ encode (MPI fpri)
 	fpri = fst $ head $ readHex fpr
 put_signature_subpacket (IssuerPacket keyid) =
-	(encode (fst $ head $ readHex keyid :: Word64), 16)
+	(encode (fst $ head $ readHex $ takeFromEnd 16 keyid :: Word64), 16)
 put_signature_subpacket (NotationDataPacket human_readable name value) =
 	(B.concat [
 		B.pack [flag1,0,0,0],
@@ -960,8 +1137,9 @@ put_signature_subpacket (FeaturesPacket supports_mdc) =
 	(B.singleton $ if supports_mdc then 0x01 else 0x00, 30)
 put_signature_subpacket (SignatureTargetPacket kalgo halgo hash) =
 	(B.concat [encode kalgo, encode halgo, hash], 31)
-put_signature_subpacket (EmbeddedSignaturePacket packet) =
-	(fst $ put_packet (assertProp isSignaturePacket packet), 32)
+put_signature_subpacket (EmbeddedSignaturePacket packet)
+	| isSignaturePacket packet = (fst $ put_packet packet, 32)
+	| otherwise = error $ "Tried to put non-SignaturePacket in EmbeddedSignaturePacket: " ++ show packet
 put_signature_subpacket (UnsupportedSignatureSubpacket tag bytes) =
 	(bytes, tag)
 
@@ -996,19 +1174,16 @@ parse_signature_subpacket 12 = do
 		sensitive = bitfield .&. 0x40 == 0x40,
 		revocation_key_algorithm = kalgo,
 		revocation_key_fingerprint =
-			pad $ map toUpper $ foldr (padB `oo` showHex) "" (B.unpack fpr)
+			pad 40 $ map toUpper $ foldr (padB `oo` showHex) "" (B.unpack fpr)
 	}
 	where
 	oo = (.) . (.)
 	padB s | odd $ length s = '0':s
 	       | otherwise = s
-	pad s = replicate (40 - length s) '0' ++ s
 -- IssuerPacket, http://tools.ietf.org/html/rfc4880#section-5.2.3.5
 parse_signature_subpacket 16 = do
 	keyid <- get :: Get Word64
-	return $ IssuerPacket (pad $ map toUpper $ showHex keyid "")
-	where
-	pad s = replicate (16 - length s) '0' ++ s
+	return $ IssuerPacket (pad 16 $ map toUpper $ showHex keyid "")
 -- NotationDataPacket, http://tools.ietf.org/html/rfc4880#section-5.2.3.16
 parse_signature_subpacket 20 = do
 	(flag1,_,_,_) <- get4word8
@@ -1076,10 +1251,7 @@ parse_signature_subpacket 31 =
 	liftM3 SignatureTargetPacket get get getRemainingByteString
 -- EmbeddedSignaturePacket, http://tools.ietf.org/html/rfc4880#section-5.2.3.26
 parse_signature_subpacket 32 =
-	fmap (EmbeddedSignaturePacket . forceSignature) (parse_packet 2)
-	where
-	forceSignature x@(SignaturePacket {}) = x
-	forceSignature _ = error "EmbeddedSignature must contain signature"
+	fmap EmbeddedSignaturePacket (parse_packet 2)
 -- Represent unsupported packets as their tag and literal bytes
 parse_signature_subpacket tag =
 	fmap (UnsupportedSignatureSubpacket tag) getRemainingByteString
@@ -1088,14 +1260,21 @@ parse_signature_subpacket tag =
 signature_issuer :: Packet -> Maybe String
 signature_issuer (SignaturePacket {hashed_subpackets = hashed,
                                    unhashed_subpackets = unhashed}) =
-	if length issuers > 0 then Just issuer else Nothing
-	where IssuerPacket issuer = issuers !! 0
-	      issuers = filter isIssuer hashed ++ filter isIssuer unhashed
-	      isIssuer (IssuerPacket {}) = True
-	      isIssuer _ = False
+	case issuers of
+		IssuerPacket issuer : _ -> Just issuer
+		_                       -> Nothing
+	where
+	issuers = filter isIssuer hashed ++ filter isIssuer unhashed
+	isIssuer (IssuerPacket {}) = True
+	isIssuer _ = False
 signature_issuer _ = Nothing
 
-find_key :: (Packet -> String) -> Message -> String -> Maybe Packet
+-- | Find a key with the given Fingerprint/KeyID
+find_key ::
+	(Packet -> String) -- ^ Extract Fingerprint/KeyID from packet
+	-> Message         -- ^ List of packets (some of which are keys)
+	-> String          -- ^ Fingerprint/KeyID to search for
+	-> Maybe Packet
 find_key fpr (Message (x@(PublicKeyPacket {}):xs)) keyid =
 	find_key' fpr x xs keyid
 find_key fpr (Message (x@(SecretKeyPacket {}):xs)) keyid =
@@ -1109,10 +1288,24 @@ find_key' fpr x xs keyid
 	| thisid == keyid = Just x
 	| otherwise = find_key fpr (Message xs) keyid
 	where
-	thisid = reverse $ take (length keyid) (reverse (fpr x))
+	thisid = takeFromEnd (length keyid) (fpr x)
+
+takeFromEnd :: Int -> String -> String
+takeFromEnd l = reverse . take l . reverse
 
 -- | SignaturePacket smart constructor
-signaturePacket :: Word8 -> Word8 -> KeyAlgorithm -> HashAlgorithm -> [SignatureSubpacket] -> [SignatureSubpacket] -> Word16 -> [MPI] -> Packet
+--
+--   <http://tools.ietf.org/html/rfc4880#section-5.2>
+signaturePacket ::
+	Word8    -- ^ Signature version (probably 4)
+	-> Word8 -- ^ Signature type <http://tools.ietf.org/html/rfc4880#section-5.2.1>
+	-> KeyAlgorithm
+	-> HashAlgorithm
+	-> [SignatureSubpacket] -- ^ Hashed subpackets (these get signed)
+	-> [SignatureSubpacket] -- ^ Unhashed subpackets (these do not get signed)
+	-> Word16 -- ^ Left 16 bits of the signed hash value
+	-> [MPI] -- ^ The raw MPIs of the signature
+	-> Packet
 signaturePacket version signature_type key_algorithm hash_algorithm hashed_subpackets unhashed_subpackets hash_head signature =
 	let p = SignaturePacket {
 		version = version,
